/*
 * Druid - a distributed column store.
 * Copyright 2012 - 2015 Metamarkets Group Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.druid.indexing.overlord;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.common.collect.ComparisonChain;
import com.google.common.util.concurrent.ListenableFuture;
import io.druid.indexing.common.TaskStatus;
import org.joda.time.DateTime;
import org.joda.time.DateTimeComparator;

/**
 * A holder for a task and different components associated with the task
 */
public class TaskRunnerWorkItem implements Comparable<TaskRunnerWorkItem>
{
  private final String taskId;
  private final ListenableFuture<TaskStatus> result;
  private final DateTime createdTime;
  private final DateTime queueInsertionTime;

  public TaskRunnerWorkItem(
      String taskId,
      ListenableFuture<TaskStatus> result
  )
  {
    this(taskId, result, new DateTime(), new DateTime());
  }

  public TaskRunnerWorkItem(
      String taskId,
      ListenableFuture<TaskStatus> result,
      DateTime createdTime,
      DateTime queueInsertionTime
  )
  {
    this.taskId = taskId;
    this.result = result;
    this.createdTime = createdTime;
    this.queueInsertionTime = queueInsertionTime;
  }

  @JsonProperty
  public String getTaskId()
  {
    return taskId;
  }

  @JsonIgnore
  public ListenableFuture<TaskStatus> getResult()
  {
    return result;
  }

  @JsonProperty
  public DateTime getCreatedTime()
  {
    return createdTime;
  }

  @JsonProperty
  public DateTime getQueueInsertionTime()
  {
    return queueInsertionTime;
  }

  public TaskRunnerWorkItem withQueueInsertionTime(DateTime time)
  {
    return new TaskRunnerWorkItem(taskId, result, createdTime, time);
  }

  @Override
  public int compareTo(TaskRunnerWorkItem taskRunnerWorkItem)
  {
    return ComparisonChain.start()
                          .compare(createdTime, taskRunnerWorkItem.getCreatedTime(), DateTimeComparator.getInstance())
                          .compare(taskId, taskRunnerWorkItem.getTaskId())
                          .result();
  }

  @Override
  public String toString()
  {
    return "TaskRunnerWorkItem{" +
           "taskId='" + taskId + '\'' +
           ", result=" + result +
           ", createdTime=" + createdTime +
           ", queueInsertionTime=" + queueInsertionTime +
           '}';
  }
}
